use std::alloc::{alloc, dealloc, Layout};

extern crate wee_alloc;

#[global_allocator]
static ALLOC: wee_alloc::WeeAlloc = wee_alloc::WeeAlloc::INIT;

// webassembly version of all the wordle logic
// because the javascript version is slow as hell on firefox

// TODO wrap all this in a struct to reduce the insane amount of
// unsafe in the code
pub static mut LOOKUP: &'static mut [u8] = &mut [];
pub static mut STRINGS: &'static mut [u8] = &mut [];
pub static mut STR_IDXS: &'static mut [*const u8] = &mut [];
pub static mut VALID_STRS: &'static mut [usize] = &mut [];

#[no_mangle]
pub fn idxs_offset() -> *const *const u8 {
    unsafe {
        STR_IDXS.as_ptr()
    }
}

fn unwrap_or_abort<T>(v: Option<T>) -> T {
    match v {
        Some(v) => v,
        None => std::process::abort(),
    }
}

unsafe fn alloc_ary<T>(sz: usize) -> &'static mut [T] {
    let layout = unwrap_or_abort(Layout::from_size_align(
            sz*std::mem::size_of::<T>(),
            std::mem::align_of::<T>()
        ).ok());
    let ptr = alloc(layout) as *mut T;
    std::slice::from_raw_parts_mut(ptr, sz)
}

unsafe fn dealloc_ary<T>(v: &'static mut [T]) {
    let layout = unwrap_or_abort(Layout::from_size_align(
            v.len()*std::mem::size_of::<T>(),
            std::mem::align_of::<T>()
        ).ok());
    dealloc(v.as_ptr() as *mut u8, layout);
}

fn realloc_ary<T>(v: &'static mut [T], sz: usize) -> &'static mut [T] {
    unsafe {
        if v.len() > 0 {
            dealloc_ary(v);
        }
        alloc_ary(sz)
    }
}


extern {
    fn fill_string(p: *mut u8);
    fn log_num_idxs(i: usize);
}

#[no_mangle]
pub extern fn init(str_sz: usize) {
    unsafe {
        STRINGS = realloc_ary(STRINGS, str_sz);
        fill_string(STRINGS.as_mut_ptr());
    }

    let num_strs = init_idx();
    unsafe {
        log_num_idxs(num_strs);

        let num_strings = STR_IDXS.len();
        LOOKUP = realloc_ary(LOOKUP, num_strings*num_strings);
    }
}

fn init_idx() -> usize {
    let strings = unsafe { &*STRINGS };

    let mut last_alpha = false;
    let mut num_words = 0;
    for v in strings.iter() {
        let cur_alpha = *v >= ('a' as u8) && *v <= ('z' as u8);
        if cur_alpha && !last_alpha {
            num_words += 1;
        }
        last_alpha = cur_alpha;
    }

    unsafe {
        if STR_IDXS.len() > 0 {
            dealloc_ary(STR_IDXS);
        }
        STR_IDXS = alloc_ary(num_words);
    }

    let idxs = unsafe { &mut *STR_IDXS };
    let mut idx = 0;
    for (i, v) in strings.iter().enumerate() {
        let cur_alpha = *v >= ('a' as u8) && *v <= ('z' as u8);
        if cur_alpha && !last_alpha {
            idxs[idx] = &strings[i] as *const u8;
            idx += 1;
        }
        last_alpha = cur_alpha;
    }

    num_words
}

static mut IDX: usize = 0;
#[no_mangle]
fn precalc(num_words: usize) {
    let mut i = unsafe { IDX };
    let idxs = unsafe { &*STR_IDXS };
    let lookup = unsafe { &mut *LOOKUP };
    let l = idxs.len();

    for _ in 0..num_words {
        if i >= idxs.len() {
            break;
        }
        let guess = idxs[i];
        for (j, reference) in idxs.iter().enumerate() {
            lookup[i*l + j] = calc_match(guess, *reference);
        }
        i += 1;
    }

    unsafe {
        IDX = i;
    }
}

#[no_mangle]
fn precalc_done() -> bool {
    unsafe {
        IDX >= STR_IDXS.len()
    }
}

#[no_mangle]
fn calc_match(guess: *const u8, reference: *const u8) -> u8 {
    let mut unmatched = vec![0u8; 26];
    let mut ret = 0;
    let mut matched = 0;

    let mut mul = 1;
    for i in 0..5 {
        unsafe {
            if *(guess.offset(i)) == *(reference.offset(i)) {
                ret += 1*mul;
                matched |= 1 << i;
            } else {
                unmatched[*(reference.offset(i)) as usize - 'a' as usize] += 1;
            }
        }
        mul *= 3;
    }

    mul = 1;
    for i in 0..5 {
        if (matched & (1 << i)) == 0 {
            let idx = unsafe { *(guess.offset(i)) as usize - ('a' as usize) };
            if unmatched[idx] > 0 {
                ret += 2*mul;
                unmatched[idx] -= 1;
            }
        }
        mul *= 3;
    }

    ret
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn it_works() {
        let result = add(2, 2);
        assert_eq!(result, 4);
    }
}
